#!/usr/bin/perl

#    travel.pl - Utility to travel in Roleplaying Games
#    Copyright (C) 2003 Michael Jung
#    e-mail: miju@phantasia.org
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$VERSION = 2.1;

use Tk;
use Tk::PNG;
use Tk::JPEG;
use Tk::NoteBook;
use Tk::Pane;

### CONSTS

%moves = ("Normal" => 1, "Free" => 0, "Forced March" => .8);
%weff = ("none" => "100");
@month = ("Nuzyael","Peonu","Kelen","Nolus","Larane","Agrazhar",
          "Azura","Halane","Savor","Ilvin","Navek","Morgat");

@t = ("spring","summer","autumn","winter");
foreach (0..11) { $season{$_} = $t[0]; shift @t if ($_%3 == 2); }

@sun = ( 1.1333, "Ula", 2.1333, "Ara", 3.1000, "Fen", 4.1333, "Ahn",
         5.1667, "Ang", 6.2333, "Nad", 7.2000, "Hir", 8.1667, "Tar",
         9.1333, "Tai", 10.1000, "Sko", 11.1000, "Mas", 12.0667, "Lad");

open(CFG,"<.travelrc") || open(CFG,"<$ENV{HOME}/.travelrc")
    || die "Can't find .travelrc";

while(<CFG>) {
    chomp;
    next if (/^[\t ]*$/ || /^\#/);
    my ($t,@t) = split(/[\t ]+/);
    $speed{eval("0x".$t[0])} = \@t if ($t eq "TRAVEL");
    $mfile = $t[0] if ($t eq "MAP");
    $weff{$t[0]} = ($t[1] =~ /(\d+)/)[0] if ($t eq "WEATHER");
    ($weathx,$weathy) = @t if ($t eq "WATCHWEATHER");
    $efile = $t[0] if ($t eq "EVENTS");
    $lefile = $t[0] if ($t eq "LOCALEVENTS");
}

close(CFG);

push @mode, grep(!/^$/,@{$speed{each %speed}}[2,4,6,8,10,12]);

### ARGV

$isfile = "-";
$osfile = "-";
$scale = 1;
$stay = 240;
$stay2 = 1440;
$moves = 1;
$elist = 10;
$weff = scalar (keys weff);

while (@ARGV) {
    my $arg = shift(@ARGV);
    if ($arg eq "-version") { die "travel.pl version $VERSION\n"; }
    elsif ($arg eq "-istate") { $isfile = shift(@ARGV); }
    elsif ($arg eq "-ostate") { $osfile = shift(@ARGV); }
    elsif ($arg eq "-trace") { $trcfile = shift(@ARGV); }
    else {
        print "Usage: travel.pl [-help] ";
        print "[-istate instate] [-ostate outstate] ";
        print "[-trace trc-file]\n";
        print "   -help = this text\n";
        print " -istate = specifies starting state file (default: stdin)\n";
	print " -ostate = specifies end state file (default: stdout)\n";
	print "  -trace = specifies trace and trace file\n";
	exit;
    }
}

### INITIALIZE GRAPHICS

$top = MainWindow->new;
$top->title("Travel");

@popt = ("padx",2,"pady",2,"fill","both","side");
@popt2 = (@popt,"left","expand","xy");

$notes = $top->NoteBook->pack(@popt2);

# Travel

$map = $notes->add("travel","-label","Travel");

$top->bind("<KeyPress-Shift_L>"=>[\&mult,0.5]);
$top->bind("<KeyRelease-Shift_L>"=>[\&mult,2]);
$top->bind("<KeyPress-Control_L>"=>[\&mult,0.25]);
$top->bind("<KeyRelease-Control_L>"=>[\&mult,4]);
$top->bind("<KeyPress-Escape>"=>[\&mult,1/6]);
$top->bind("<KeyRelease-Escape>"=>[\&mult,6]);

$bframe = $map->Frame->pack(@popt,"left"); # Stuff
$cframe = $map->Frame->pack(@popt2); # Map

# Buttons

@bhash = ("relief","groove","borderwidth",2);
@bhash1 = (@bhash,"textvariable");
@bopt = (@popt,"top","expand","xy");
@infonly = ("state","disabled","disabledforeground","grey50");

$bframe->Button(@bhash,"text","Exit","command",\&save_end)->pack(@bopt);
$bframe->Button(@bhash1,\$now,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$dist,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$mov,"command",\&reset)->pack(@bopt);
$bframe->Button(@bhash1,\$fstay,"command",\&move)->pack(@bopt);
$bfree = $bframe->Button(@bhash,"command",\&tog_free)->pack(@bopt);
$btrace = $bframe->Button(@bhash1,\$trace,"command",\&tog_trace)->pack(@bopt);
$bterrain = $bframe->Button(@bhash,@infonly)->pack(@bopt);
$bframe->Button(@bhash,"text","Mode","command",\&tog_modes)->pack(@bopt);
$bmrat = $bframe->Button(@bhash,"command",\&tog_mvrate)->pack(@bopt);
$bframe->Entry("textvariable",\$speed)->pack(@bopt);

# Canvas

$big = $cframe->Canvas(@bhash)->pack(@popt2);
$big->bind("all","<ButtonRelease-1>"=>[\&move, Ev("x"), Ev("y")]);
$big->bind("all","<ButtonRelease-3>"=>[\&calc_estim, Ev("x"), Ev("y")]);
$big->Tk::bind("<Visibility>"=>\&mark);
$big->Tk::bind("<Expose>"=>\&mark);

# Weather

$weath = $notes->add("weather","-label","Weather");

$bframe = $weath->Frame->pack(@popt,"left"); # Stuff
$cframe = $weath->Frame->pack(@popt2); # Weather

$bframe->Button(@bhash,"text","Exit","command",\&save_end)->pack(@bopt);
$bframe->Button(@bhash1,\$now,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$dist,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$fstay2,"command",\&stay)->pack(@bopt);
$bframe->Button(@bhash1,\$sun,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$month,@infonly)->pack(@bopt);
$bframe->Button(@bhash,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$fstay,"command",\&move)->pack(@bopt);
$bframe->Button(@bhash1,\$dawn,"command",\&dawn)->pack(@bopt);
$bframe->Button(@bhash1,\$moon,@infonly)->pack(@bopt);
$beff = $bframe->Button(@bhash,"command",\&tog_eff)->pack(@bopt);

@copt = ("-scrollbars","osoe");
$canv = $cframe->Scrolled("Canvas",@copt)->pack(@popt2);

$canv->bind("all","<ButtonRelease-1>"=>[\&stay, Ev("x"), Ev("y")]);
$canv->bind("all","<ButtonRelease-3>"=>[\&stay_estim, Ev("x"), Ev("y")]);

# Events

$evts = $notes->add("events","-label","Events");

@popt3 = ("padx",0,"pady",0,"fill","both","side","left","expand","xy");

$bframe = $evts->Frame->pack(@popt,"left"); # Stuff
$pane = $evts->Scrolled("Pane",@copt,"-sticky","nsew")->pack(@popt2)
    ->Frame->pack(@bopt);

$bframe->Button(@bhash,"text","Exit","command",\&save_end)->pack(@bopt);
$bframe->Button(@bhash1,\$now,@infonly)->pack(@bopt);
$bframe->Button(@bhash,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$fstay2,"command",\&stay)->pack(@bopt);
$bframe->Button(@bhash1,\$moon,@infonly)->pack(@bopt);
$bframe->Button(@bhash1,\$fstay,"command",\&move)->pack(@bopt);
$bframe->Button(@bhash1,\$month,@infonly)->pack(@bopt);
$bframe->Button(@bhash,"text","Add","command",&edit_ev(-1))->pack(@bopt);
$bframe->Button(@bhash1,\$last,@infonly)->pack(@bopt);
$bframe->Button(@bhash,"text","Roll","command",\&roll)->pack(@bopt);
$bframe2 = $bframe->Frame->pack(@bopt);
$bframe2->Button(@bhash,"text","<","command",\sub{&events(-1)})->pack(@popt3);
$bframe2->Button(@bhash,@infonly,"textvariable",\$elist)->pack(@popt3);
$bframe2->Button(@bhash,"text",">","command",\sub{&events(1)})->pack(@popt3);

### INITIALIZE POSITION

open(INSTATE,"<$isfile") || error($isfile);
(($datw,$datd,$datm,$daty) = split(/\./,<INSTATE>))
    || print STDERR "Date missing!\n";
(($bcx,$bcy) = split(/,/,<INSTATE>)) || print STDERR "Position missing!\n";
$wpart = <INSTATE> || print STDERR "Remainder of watch missing!\n";
chomp $daty;
chomp $bcy;
chomp $wpart;
$wleg = 0;

while (<INSTATE>) { push @weather, split(/,/,$_); }

close(INSTATE);
open(OUTSTATE,">$osfile") || error($osfile);
open(TRC,">>$trcfile") || error($trcfile) if (defined($trcfile));

### INITIALIZE EVENTS

&load_ev("g",$efile);
&load_ev("l",$lefile);

### INITIALIZE BOARD

$mx = $big->cget("width")/2;
$my = $big->cget("height")/2;

$totpic = $top->Photo("-file", $mfile);
$pic = $top->Photo("-width",0,"-height",0);
&picmove;

$img = $big->createImage(0,0,"image",$pic,"anchor","nw");
$bm = $top->Bitmap("-file","mark.xbm");
$mark = $big->createImage($mx,$my,"image",$bm,"anchor","c");

$canv->configure("scrollregion", [0,0,6*$weathx,10*$weathy]);

### INITIALIZE PARAMETERS

@cmode = ("foot");
&reset(0);
&mult(0,1);
&tog_free;
&tog_trace;
&tog_mvrate;
&tog_eff;
&calc_estim(0,$mx,$my);
&calc_time;
&events(0);

### MAIN

MainLoop;

### SUBROUTINES

sub move {
    my @num = (@_) ? &calc_estim(@_) : ($stay,0,0);
    
    $wpart += $num[0];
    $wleg = (@_) ? $wleg + $num[0] : 0;
    &reset($wleg);

    $bcx += int $num[1]; 
    $bcy += int $num[2];
    &picmove;
    print(TRC "$bcx,$bcy\n") if ($trcfile && $btrace->cget("text") eq "Trace");
    &calc_time;
}

sub load_ev {
    my $l = shift;
    my $f = shift;
    open(EVENTS,"<$f") || error($f);
    while (<EVENTS>) {
        my @e;
        if (/^\#(Last roll: \d+\.\d+ \(\d+\))/) {
            $last = "$1";
        }
        elsif (/(\d+)(?:\.(\d+)\s*(?:\((\d)\)\s*(?:(\d)\:(\d+))?)?)?\s+(.*)/) {
            @e = ($6,$2,$1,$3,$4*60+$5);
            @e = ($6,$1) if ($2 eq "");
        }
        push(@{$ev{sprintf("%02d.%02d.%01d.%03d",@e[1..4])}},[$l,$e[0]])
            if (defined($e[0]));
    }
    close(EVENTS);
}

sub roll {
    my ($d,$m,$w) = ($last =~ /(\d+)\.(\d+) \((\d)/);
    foreach my $i (1..60) {
        my $ww = ($w + $i - 1)%6 + 1;
        my $dd = ($d + int(($i + $w - 1)/6) - 1)%30 + 1;
        my $mm = ($m + ($dd < $d ? 0 : -1))%12 + 1;
        my $r = d(20);
        my $t;
	$t .= "U" if ($r >= 16); $t .= "u" if ($r >= 19);
	$t .= "R" if ($r >= 14); $t .= "r" if ($r >= 19);
	$t .= "H" if ($r >= 17); $t .= "h" if ($r >= 19);
	$t .= "W" if ($r >= 19); $t .= "w" if ($r >= 20);
	$t .= "Ss" if ($r >= 19); $t .= "Oo" if ($r >= 20);
        push
            (@{$ev{sprintf("%02d.%02d.%1d.%03d",$mm,$dd,$ww,d(240)-1)}},
             ["l","Random: $t"])
            if ($t ne "");
        $last = "Last roll: $dd.$mm ($ww)";
    }
    &events(0);
}

sub events {
    my $lelist = $elist += $_[0];
    foreach ($lelist..$#bev) { (shift @bev)->destroy(); }
    foreach ($#bev+1..$lelist-1) {
        push(@bev, my $f = $pane->Frame(@bhash)->pack("fill","x"));
        $f->Label("anchor","c","width",6)->pack(@popt,"left");
        $f->Label("anchor","c","width",14)->pack(@popt,"left");
        my $l = $f->Label("anchor","nw")->pack(@popt,"left","expand","x");
        $l->bind("<ButtonRelease-1>"=>&move_ev($_));
        $l->bind("<ButtonRelease-3>"=>&edit_ev($_));
        
    }
    my ($m,$d,$w) = ($datm,$datd,$datw);
    while ($lelist > 0) {
        my $ds = "%02d.00.0.000";
        my $ev = sprintf($ds,$m);
        my $del = ($m-$datm)*30*6*240;
        foreach (@{$ev{$ev}}) {
            &put_event(--$lelist, $del, "%s", \$_, $m)
                if ($lelist > 0 && $del >= 0 && $_ ne "");
        }
        $ds = "%02d.%02d.0.000";
        $d = 1 if ($d == 0);
        while ($d != 0) {
            my $ev = sprintf($ds,$m,$d);
            my $del = (($m-$datm)*30 + ($d-$datd))*6*240;
            foreach (@{$ev{$ev}}) {
                &put_event(--$lelist, $del, "%s.%s.", \$_, $d, $m)
                    if ($lelist > 0 && $del >= 0 && $_ ne "");
            }
            $ds = "%02d.%02d.%01d.000";
            $w = 1 if ($w == 0);
            while ($w != 0) {
                my $ev = sprintf($ds,$m,$d,$w);
                my $del = ((($m-$datm)*30 + ($d-$datd))*6 + ($w-$datw))*240;
                foreach (@{$ev{$ev}}) {
                    &put_event
                        (--$lelist, $del, "%2s.%2s (%1d)", \$_, $d, $m, $w)
                        if ($lelist > 0 && $del >= 0 && $_ ne "");
                }
                my $t = substr($ev,0,-3);
                foreach (sort grep !/000/, grep /$t/, keys %ev) {
                    my $p = int substr($_,-3);
                    my $del = ((($m-$datm)*30 + ($d-$datd))*6 +
                               ($w-$datw))*240 + ($p-$wpart);
                    foreach (@{$ev{$_}}) {
                        &put_event
                            (--$lelist, $del, "%2s.%2s (%1d) %1d:%02d", \$_,
                             $d, $m, $w, $p/60, $p%60)
                            if ($lelist > 0 && $del >= 0 && $_ ne "");
                    }
                }
                $w = ($w + 1)%7;
            }
            $d = ($d + 1)%31;
        }
	$m = ($m + 1)%12;
    }
}

sub put_event {
    my ($i, $del,$ds,$ct) = (shift @_, shift @_, shift @_, shift @_);
    my ($sdel,$txt) = ("",sprintf($ds,@_));
    if ($del < 60*10) {
        $sdel = ($del != 0) ? &format($del) : "-";
    }
    elsif ($del < 10*24*60) {
        $sdel = int($del/(60*24))."d " if ($del/(24*60) >= 1);
        $sdel .= int(($del%(24*60))/60)."h" if ($del%(24*60) >= 60);
    }
    else {
        $sdel = int($del/(60*24))."d";
    }
    $tev0[$#bev - $i] = $del;
    $tev1[$#bev - $i] = \${$$ct}[1];
    my @c = $bev[$#bev - $i]->children;
    $c[0]->configure("text",$sdel);
    $c[1]->configure("text",$txt);
    $c[2]->configure("textvariable",$tev1[$#bev - $i]);
}

sub calc_estim {
    &speed(@_[1..2]);
    if ($speed * $scale == 0) {
	print STDERR "Scale and Speed must not be zero!";
	return (0,0,0);
    }
    my $dx = $_[1] - $mx;
    my $dy = $_[2] - $my;
    my $t = int (240 * sqrt($dx**2 + $dy**2)/($scale*$speed));
    $t *= $moves{$bfree->cget("text")};
    $t /= ($weff{$beff->cget("text")}/100);
    $dist = "Distance:".&format($t);
    return ($t, $dx, $dy);
}

sub calc_time {
    while ($wpart >= 240) {
	$wpart -= 240;
	$datw = ($datw%6) + 1;
        if ($datw == 1) {
            $datd = ($datd%30) + 1;
            shift @weather foreach (0..5);
        }
	$datm = ($datm%12) + 1 if ($datw*$datd == 1);
	$daty++ if ($datw*$datd*$datm == 1);
    }
    $now = sprintf
        ("%d.%d.%d (%d) ".&format($wpart),$datd,$datm,$daty,$datw);
    my $t = "new moon";
    $t = "waxing: ".int(100*($datd-1)/15)."%" if ($datd > 1);
    $t = "full moon" if ($datd == 16);
    $t = "waning: ".int(100*(31-$datd)/15)."%" if ($datd > 16);
    $moon = $t;
    $sun = &sunsign;
    &calc_weather;
    $month = $month[$datm-1];
    &events(0);
}

sub save_end {
    print OUTSTATE "$datw.$datd.$datm.$daty\n$bcx,$bcy\n${wpart}\n";
    print OUTSTATE join(",", @weather),"\n";
    close(OUTSTATE);
    close(TRC);
    open(EVENT, ">$lefile");
    foreach my $k (sort keys %ev) {
        foreach (@{$ev{$k}}) {
            if (${$_}[0] eq "l" && ${$_}[1] ne "") {
                my ($m,$d,$w,$p) = ($k =~ /(\d+).(\d+).(\d+).(\d+)/);
                print EVENT 
                    ($d != 0 ? "$d.$m" : "$m"), ($w != 0 ? " ($w) " : " "),
                    ($p != 0 ? &format($p) : ""), " ${$_}[1]\n";
            }
        }
    }
    print EVENT "#$last\n";
    close(EVENT);
    exit;
}

sub mark {
    my ($x,$y) = ($big->width/2,$big->height/2);
    $big->move($mark,$x-$mx,$y-$my);
    ($mx,$my) = ($x,$y);
    &picmove;
}

sub toggle {
    local *x = shift;
    my @t = (keys %x);
    $x = ($x+1)%($#t+1);
    $_[0]->configure("text",$t[$x]);
}

sub edit_ev {
    my $i = shift;
    return \sub {
        if (!defined($dev)) {
            $dev = $top->Toplevel;
            my $f0 = $dev->Frame->pack(@popt,"top");
            if ($i != -1) {
                $f0->Entry("textvariable",$tev1[$i])->pack(@bopt);
                $dev->Button(@bhash,"text","ok","command",\&ev_end)->pack
                    (@bopt);
            }
            else {
                my $f1 = $dev->Frame->pack(@popt,"top");
                my ($d,$e);
                $f0->Label("text","Date (mm.dd.w.mmm):")->pack(@popt,"left");
                $f0->Entry("textvariable",\$d)->pack(@popt,"left");
                $f1->Label("text","Event:")->pack(@popt,"left");
                $f1->Entry("textvariable",\$e)->pack
                    (@popt,"left","expand","x");
                $dev->Button(@bhash,"text","ok","command",&ev_end2(\$d,\$e))
                    ->pack(@bopt);
            }
        }
    }
}
        
sub tog_modes {
    if (!defined($dmod)) {
        $dmod = $top->Toplevel;
        my $f0 = $dmod->Frame->pack(@popt,"top",);
        my $f1 = $f0->Frame(@bhash)->pack(@popt,"top");
        my $f2 = $f0->Frame->pack(@popt,"top");
        foreach my $i (@mode) {
            $bmod{$i} = $f1->Button
                (@bhash,"text",$i,"command",\sub{&tog_mode($i)})->pack(@bopt);
            $bmod{$i}->configure("relief","sunken")
                if (grep(/$i/, @cmode) > 0);
        }
        $f2->Button(@bhash,"text","Close","command",\&mode_end)->pack(@bopt);
    }
}

sub tog_mode {
    if ($bmod{$_[0]}->cget("relief") eq "groove") {
        $bmod{$_[0]}->configure("relief","sunken");
        push @cmode, $_[0];
    }
    else {
        $bmod{$_[0]}->configure("relief","groove");
        @cmode = grep !/$_[0]/, @cmode;
    }
}

sub tog_mvrate {
    $bmrat->configure
        ("text",$bmrat->cget("text") ne "Auto" ? "Auto" : "Manual ($aspeed)" );
}

sub mult {
    $fstay = "Stay (".&format($stay *= $_[1]).")";
    $fstay2 = "Stay (".(($stay2 /= $_[1])/1440)." day".($stay2==1440?")":"s)");
}

sub speed {
    my (@col) = $pic->get(@_);
    my $col = $col[2] + ($col[1]<<8) + ($col[0]<<16);
    my (%c,%s);
    foreach my $j (@cmode) {
        foreach my $i (keys %speed) {
            my %mods = @{$speed{$i}};
            my $sp = $mods{$j};
            if ($i == $col) {
                if ($sp =~ /(\d+)%/) {
                    $sp = ($1)/100 * &pure_speed($j,$_[0],$_[1]);
                }
                $c{$j} = $sp;
                $s{$j} = -$i;
                last;
            }
            if ($sp !~ /%/) {
                my $d = abs($col[2] - ($i & 0xff)) +
                    abs($col[1] - (($i & 0xff00)>>8)) +
                    abs($col[0] - (($i & 0xff0000)>>16));
                $s{$j} += 1/$d;
                $c{$j} += $sp/$d;
            }
        }
        $c{$j} = int($c{$j}/($s{$j} > 0 ? $s{$j} : 1));
    }
    my $c = $cmode[0];
    $aspeed = $c{$c};
    foreach my $j (@cmode) { $aspeed = min($c{$j},$aspeed); }

    $bterrain->configure
        ("text", ($s{$c} > 0 ? "Guesswork" : ${$speed{-$s{$c}}}[1]));

    if ($bmrat->cget("text") eq "Auto") {
        $speed = $aspeed;
    }
    else {
        $bmrat->configure("text", "Manual ($aspeed)");
    }
}

sub pure_speed {
    my @dir = ([-1,-1],[0,-1],[1,-1],[-1,0],[1,0],[-1,1],[0,1],[1,1]); 
    my $f = 1;
    while ($f > 0) {
        $f++;
        my @col;
        foreach my $i (@dir) {
            my ($x,$y) = ($_[1] + $f*${$i}[0], $_[2] + $f*${$i}[1]);
            if ($x >=0 && $y >= 0 && $x <= $pic->width && $y <= $pic->height) {
                @col = $pic->get($x,$y);
            }
        }
        my $col = $col[2] + ($col[1]<<8) + ($col[0]<<16);
        if (defined($speed{$col})) {
            my %m = @{$speed{$col}};
            $f = -$m{$_[0]}-1 if ($m{$_[0]} !~ /%/);
        }
    }
    return -$f-1;
}

sub calc_weather {
    local *w = *weather;
    $w[0] = d(20) if (!@w);
    push (@w, ($w[$#w] + &weather($w[$#w]) - 1)%20 + 1) while ($#w < 59);

    $canv->delete("all");
    foreach my $i (0..9) {

        # Night/Day
        my $y = $weathy*$i;
        my $off = - sin(($datd + $i + 30*$datm - 31)*atan2(1,1)/45);
        my $r = $canv->createRectangle
            (1, $y + 1, 6*$weathx - 2, $y + $weathy - 2,
             "fill", "white", "outline", "white");
        my $r = $canv->createRectangle
            (1, $y + 1, (3 + $off)*$weathx/2 - 2, $y + $weathy - 2,
             "fill", "grey70", "outline", "grey70");
        $r = $canv->createRectangle
            (6*$weathx - 1, $y + 1, (9 - $off)*$weathx/2 - 2, $y + $weathy - 2,
             "fill", "grey70", "outline", "grey70");

        # Mark
        if ($i == 0) {
            my $x = (240*$datw + $wpart - 241)/240 * $weathx;
            $r = $canv->createRectangle
                ($x - 5, 1, $x + 5, $weathy - 2,
                 "fill", "grey40", "outline", "grey40");
            # Dawn/Dusk
            my $d = (3 + $off)/2 - $x/$weathx;
            if ($d > .005) {
                $dawn = "Dawn: ".&format($d*240);
            }
            else {
                $d = (9 - $off)/2 - $x/$weathx;
                if ($d > .005) {
                    $dawn = "Dusk: ".&format($d*240);
                }
                else {
                    $d = (15 + $off)/2 - $x/$weathx;
                    $dawn = "Dawn: ".&format($d*240);
                }
            }
        }

        # Phenomena
        my $s = $season{(int(($datd + $i - 1)/30) + $datm - 1)%12};
        foreach my $j (0..5) {
            my $p = $top->Bitmap
                ("-file","weather/${s}_".($weather[6*$i+$j]).".xbm");
            $canv->createImage($weathx*$j, $y, "image", $p, "anchor", "nw");
        }
    }
}

sub stay_estim {
    my @xy = $canv->coords("current");
    if (@xy) {
        my $t = 3*480*$xy[1]/$weathy;
        $t += 240*$_[1]/$weathx;
        $t -= $wpart + ($datw - 1)*240;
        $dist = "Distance:".(($t > 0) ? &format($t) : "-".&format(-$t));
        return $t if ($t > 0);
    }
}

sub sunsign {
    my ($i,$j,$k);
    $i++ while ($i*2 < $#sun && $datm + $datd/30 + .005 >= $sun[$i*2]);
    $j++ while ($j*2 < $#sun && $datm + $datd/30 + .083>= $sun[$j*2]);
    $k++ while ($k*2 < $#sun && $datm + $datd/30 - .050 >= $sun[$k*2]);
    my $ms = $sun[($i*2 + $#sun)%($#sun+1)];
    $ms .= "/".$sun[($j*2 + $#sun)%($#sun+1)] if ($i != $j);
    $ms .= "/".$sun[($k*2 + $#sun)%($#sun+1)] if ($i != $k);
    return $ms;
}

sub picmove {
    my $cpx1 = max(0,$bcx - $mx);
    my $cpx2 = min($bcx + $mx,$totpic->width);
    my $mvx = max(0,$mx - $bcx);
    my $szx = 2*$mx - max(0,$mx + $bcx - $totpic->width);

    my $cpy1 = max(0,$bcy - $my);
    my $mvy = max(0,$my - $bcy);
    my $cpy2 = min($bcy + $my,$totpic->height);
    my $szy = 2*$my - max(0,$my + $bcy - $totpic->height);

    $pic->configure("width",$szx + 1);
    $pic->configure("height",$szy + 1);
    $pic->copy($totpic, "-from", $cpx1, $cpy1, $cpx2, $cpy2);
    $big->coords($img, $mvx, $mvy);
}

sub ev_end2 {
    my ($d,$e) = @_;
    return \sub { push @{$ev{$$d}}, ["l", $$e]; &ev_end }
}

sub dawn { my @t = ($dawn =~ /(\d+):(\d+)/); &move2($t[0]*60 + $t[1]); }

sub reset { $wleg = 0 if (!@_); $mov = "Move: ".&format($wleg); }

sub min { return ($_[0] < $_[1]) ? $_[0] : $_[1]; }

sub max { return ($_[0] > $_[1]) ? $_[0] : $_[1]; }

sub d { return 1 + int rand $_[0]; }

sub stay { &move2((@_) ? &stay_estim(@_) : $stay2); }

sub weather { return 0 if (d(10) < 7); return -int((d(4)**2 - 6)/5); }

sub mode_end { $dmod->destroy; undef $dmod; }

sub ev_end { &events(0); $dev->destroy; undef $dev; }

sub format { return sprintf("%d:%02d",$_[0]/60,$_[0]%60); }

sub error { print STDERR "Unable to open ".$_[0]."!\n"; }

sub move2 { my $s = $stay; $stay = $_[0]; &move(); $stay = $s; }

sub tog_free { toggle(*moves,$bfree); }

sub tog_eff { toggle(*weff,$beff); }

sub tog_trace { $trace = ($trace eq "Trace") ? "No trace" : "Trace"; }

sub move_ev { my $i = shift; return \sub { &move2($tev0[$i]); } }
